#include <e.h>
#include "e_kbd_dict.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

static void
_e_kbd_dict_lookup_build_line(E_Kbd_Dict *kd, const char *p, const char *eol, 
			      int *glyphs)
{
   char *s;
   int p2;
   
   s = alloca(eol - p + 1);
   strncpy(s, p, eol - p);
   s[eol - p] = 0;
   p2 = evas_string_char_next_get(s, 0, &(glyphs[0]));
   if ((p2 > 0) && (glyphs[0] > 0))
     p2 = evas_string_char_next_get(s, p2, &(glyphs[1]));
}

static void
_e_kbd_dict_lookup_build(E_Kbd_Dict *kd)
{
   const char *p, *e, *eol;
   int glyphs[2], pglyphs[2];

   p = kd->file.dict;
   e = p + kd->file.size;
   pglyphs[0] = pglyphs[1] = 0;
   while (p < e)
     {
	eol = strchr(p, '\n');
	if (!eol) break;
	if (eol > p)
	  {
	     glyphs[0] = glyphs[1] = 0;
	     _e_kbd_dict_lookup_build_line(kd, p, eol, glyphs);
	     if ((glyphs[1] != pglyphs[1]) ||
		 (glyphs[0] != pglyphs[0]))
	       {
		  int v1, v2;
		  
		  if (isspace(glyphs[0]))
		    {
		       glyphs[0] = 0;
		       glyphs[1] = 0;
		    }
		  else if (isspace(glyphs[1]))
		    glyphs[1] = 0;
		  v1 = tolower(glyphs[0]) & 0xff;
		  v2 = tolower(glyphs[1]) & 0xff;
		  kd->lookup.tuples[v1][v2] = p;
		  pglyphs[0] = v1;
		  pglyphs[1] = v2;
	       }
	  }
	p = eol + 1;
     }
}

static int
_e_kbd_dict_open(E_Kbd_Dict *kd)
{
   struct stat st;
   
   kd->file.fd = open(kd->file.file, O_RDONLY);
   if (kd->file.fd < 0)
     return 0;
   if (fstat(kd->file.fd, &st) < 0)
     {
	close(kd->file.fd);
	return 0;
     }
   kd->file.size = st.st_size;
   kd->file.dict = mmap(NULL, kd->file.size, PROT_READ, MAP_SHARED,
			kd->file.fd, 0);
   if ((kd->file.dict== MAP_FAILED) || (kd->file.dict == NULL))
     {
	close(kd->file.fd);
	return 0;
     }
   return 1;
}

static void
_e_kbd_dict_close(E_Kbd_Dict *kd)
{
   memset(kd->lookup.tuples, 0, sizeof(kd->lookup.tuples));
   munmap((void *)kd->file.dict, kd->file.size);
   close(kd->file.fd);
}

EAPI E_Kbd_Dict *
e_kbd_dict_new(const char *file)
{
   // alloc and load new dict - build quick-lookup table. words MUST be sorted
   E_Kbd_Dict *kd;
   
   kd = E_NEW(E_Kbd_Dict, 1);
   if (!kd) return NULL;
   kd->file.file = evas_stringshare_add(file);
   if (!kd->file.file)
     {
	free(kd);
	return NULL;
     }
   if (!_e_kbd_dict_open(kd))
     {
	evas_stringshare_del(kd->file.file);
	free(kd);
	return NULL;
     }
   _e_kbd_dict_lookup_build(kd);
   return kd;
}

EAPI void
e_kbd_dict_free(E_Kbd_Dict *kd)
{
   // free dict and anything in it
   e_kbd_dict_word_letter_clear(kd);
   // FIXME: free pending writes 
   // FIXME: free matches
   _e_kbd_dict_close(kd);
   free(kd);
}

EAPI void
e_kbd_dict_save(E_Kbd_Dict *kd)
{
   // save any changes (new words added, usage adjustments).
   // all words MUST be sorted
   // NOTE: should really save every N words that end up in the changes buffer
   // FIXME: sort changed.writes
   // FIXME: open new file for writes
   // FIXME: walk source, for each, if line word is in writes, replace with
   // writes, or if writes char fits between 2 source lines, insert. if
   // in writes, usage == -1, then dont write word out
   _e_kbd_dict_close(kd);
   // FIXME: unlink old file
   // FIXME: mv new file to old
   _e_kbd_dict_open(kd);
}

static E_Kbd_Dict_Word *
_e_kbd_dict_changed_write_find(E_Kbd_Dict *kbd, const char *word)
{
   Evas_List *l;
   
   for (l = kbd->changed.writes; l; l = l->next)
     {
	E_Kbd_Dict_Word *kw;
	
	kw = l->data;
	if (!strcmp(kw->word, word)) return kw;
     }
   return NULL;
}

static void
_e_kbd_dict_changed_write_add(E_Kbd_Dict *kbd, const char *word, int usage)
{
   E_Kbd_Dict_Word *kw;
   
   kw = E_NEW(E_Kbd_Dict_Word, 1);
   kw->word = evas_stringshare_add(word);
   kw->usage = usage;
   kbd->changed.writes = evas_list_prepend(kbd->changed.writes, kw);
}

static const char *
_e_kbd_dict_find(E_Kbd_Dict *kd, const char *word)
{
   const char *p, *e, *eol;
   int glyphs[2], p2, v1, v2, len;

   p = kd->file.dict;
   e = p + kd->file.size;
   glyphs[0] = glyphs[1] = 0;
   p2 = evas_string_char_next_get(word, 0, &(glyphs[0]));
   if ((p2 > 0) && (glyphs[0] > 0))
     p2 = evas_string_char_next_get(word, p2, &(glyphs[1]));
   v1 = tolower(glyphs[0]) & 0xff;
   v2 = tolower(glyphs[1]) & 0xff;
   p = kd->lookup.tuples[v1][v2];
   printf("lookup %s = %p\n", word, p);
   if (!p) return NULL;
   len = strlen(word);
   while (p < (e - len))
     {
	eol = strchr(p, '\n');
	if (!eol) break;
	if (eol > p)
	  {
	     if (!strncasecmp(p, word, len)) return p;
	  }
	if (strncasecmp(p, word, p2)) return NULL;
        p = eol + 1;
     }
   return NULL;
}

EAPI void
e_kbd_dict_word_usage_adjust(E_Kbd_Dict *kd, const char *word, int adjust)
{
   // add "adjust" to word usage count
   E_Kbd_Dict_Word *kw;
   
   kw = _e_kbd_dict_changed_write_find(kd, word);
   printf("find write = %p\n", kw);
   if (kw)
     kw->usage += adjust;
   else
     {
	const char *line;
	int usage = 0;
	
	line = _e_kbd_dict_find(kd, word);
	if (line)
	  {
	     const char *spc, *p;
	     
	     spc = NULL;
	     for (p = line; p[0] != '\n'; p++)
	       {
		  if (isspace(*p))
		    {
		       spc = p;
		       break;
		    }
	       }
	     if (spc)
	       {
		  spc++;
		  if (isdigit(spc[0])) usage = atoi(spc);
	       }
	     printf("word %s, line %p, usage == %i\n", word, line, usage);
	  }
	usage += adjust;
	_e_kbd_dict_changed_write_add(kd, word, usage);
     }
}

EAPI void
e_kbd_dict_word_delete(E_Kbd_Dict *kd, const char *word)
{
   // delete a word from the dictionary
   E_Kbd_Dict_Word *kw;
   
   kw = _e_kbd_dict_changed_write_find(kd, word);
   if (kw)
     kw->usage = -1;
   else
     {
	if (_e_kbd_dict_find(kd, word))
	  _e_kbd_dict_changed_write_add(kd, word, -1);
     }
}

EAPI void
e_kbd_dict_word_letter_clear(E_Kbd_Dict *kd)
{
   // clear the current word buffer
   while (kd->word.letters)
     e_kbd_dict_word_letter_delete(kd);
}

EAPI void
e_kbd_dict_word_letter_add(E_Kbd_Dict *kd, const char *letter, int dist)
{
   // add a letter with a distance (0 == closest) as an option for the current
   // letter position - advance starts a new letter position
   Evas_List *l, *list;
   E_Kbd_Dict_Letter *kl;
   
   l = evas_list_last(kd->word.letters);
   if (!l) return;
   list = l->data;
   kl = E_NEW(E_Kbd_Dict_Letter, 1);
   if (!kl) return;
   kl->letter = evas_stringshare_add(letter);
   kl->dist = dist;
   list = evas_list_append(list, kl);
   l->data = list;
}

EAPI void
e_kbd_dict_word_letter_advance(E_Kbd_Dict *kd)
{
   // start a new letter in the word
   // FIXME:
   kd->word.letters = evas_list_append(kd->word.letters, NULL);
}

EAPI void
e_kbd_dict_word_letter_delete(E_Kbd_Dict *kd)
{
   // delete the current letter completely
   // FIXME:
   Evas_List *l, *list;
   
   l = evas_list_last(kd->word.letters);
   if (!l) return;
   list = l->data;
   while (list)
     {
	E_Kbd_Dict_Letter *kl;
	
	kl = list->data;
	evas_stringshare_del(kl->letter);
	free(kl);
	list = evas_list_remove_list(list, list);
     }
   kd->word.letters = evas_list_remove_list(kd->word.letters, l);
}

static void
_e_kbd_dict_matches_lookup_iter(E_Kbd_Dict *kd, Evas_List *word, Evas_List *more)
{
   
}

EAPI void
e_kbd_dict_matches_lookup(E_Kbd_Dict *kd)
{
   // find all matches and sort them
   // FIXME:
   Evas_List *l, *l2, *list, *word;
   
   for (l = kd->word.letters; l; l = l->next)
     {
	list = l->data;
	for (l2 = list; l2; l2 = l2->next)
	  {
	     E_Kbd_Dict_Letter *kl;
	     
	     kl = l2->data;
	  }
     }
}

EAPI void
e_kbd_dict_matches_first(E_Kbd_Dict *kd)
{
   // jump to first match
   // FIXME:
}

EAPI void
e_kbd_dict_matches_next(E_Kbd_Dict *kd)
{
   // jump to next match
   // FIXME:
}

EAPI const char *
e_kbd_dict_matches_match_get(E_Kbd_Dict *kd, int *pri_ret)
{
   // return the word (string utf-8) for the current match
   // FIXME:
}
